import cv2
import numpy as np


def pendulum_render_fn(img, observation, action):
    height, width, _ = img.shape
    side_length = min(width, height)
    state = observation.msgs[-1].data

    img += 255
    length = side_length // 3
    sin_theta, cos_theta = np.sin(state[0]), np.cos(state[0])

    # Draw pendulum
    img = cv2.line(
        img,
        (width // 2, height // 2),
        (width // 2 + int(length * sin_theta), height // 2 - int(length * cos_theta)),
        (0, 0, 255),
        max(side_length // 32, 1),
    )

    # Draw mass
    img = cv2.circle(
        img, (width // 2 + int(length * sin_theta), height // 2 - int(length * cos_theta)), side_length // 16, (0, 0, 0), -1
    )

    # Draw velocity vector
    img = cv2.arrowedLine(
        img,
        (width // 2 + int(length * sin_theta), height // 2 - int(length * cos_theta)),
        (
            width // 2 + int(length * (sin_theta + state[1] * cos_theta / 5)),
            height // 2 + int(length * (-cos_theta + state[1] * sin_theta / 5)),
        ),
        (0, 0, 255),
        max(side_length // 240, 1),
    )

    # Draw joint
    img = cv2.circle(img, (width // 2, height // 2), max(side_length // 160, 1), (0, 0, 0), -1)

    return img


def disc_pendulum_render_fn(img, observation, action):
    height, width, _ = img.shape
    side_length = min(width, height)
    state = observation.msgs[-1].data

    img += 255
    length = 2 * side_length // 9
    sin_theta, cos_theta = np.sin(state[0]), np.cos(state[0])

    img = cv2.circle(img, (width // 2, height // 2), side_length // 3, (255, 0, 0), -1)
    img = cv2.circle(img, (width // 2, height // 2), side_length // 12, (192, 192, 192), -1)
    img = cv2.circle(
        img,
        (width // 2 + int(length * sin_theta), height // 2 - int(length * cos_theta)),
        side_length // 9,
        (192, 192, 192),
        -1,
    )

    # Draw velocity vector
    img = cv2.arrowedLine(
        img,
        (width // 2 + int(length * sin_theta), height // 2 - int(length * cos_theta)),
        (
            width // 2 + int(length * (sin_theta + state[1] * cos_theta / 5)),
            height // 2 + int(length * (-cos_theta + state[1] * sin_theta / 5)),
        ),
        (0, 0, 0),
        2,
    )
    return img



import eagerx
from eagerx.core.specs import EngineStateSpec
from typing import Any


class DummyState(eagerx.EngineState):
    @classmethod
    def make(cls) -> EngineStateSpec:
        return cls.get_specification()

    def initialize(self, spec: EngineStateSpec, simulator: Any):
        pass

    def reset(self, state: Any):
        pass


class SetGymAttribute(eagerx.EngineState):
    @classmethod
    def make(cls, attribute: str) -> EngineStateSpec:
        spec = cls.get_specification()
        spec.config.attribute = attribute
        return spec

    def initialize(self, spec: EngineStateSpec, simulator: Any):
        self.attribute = spec.config.attribute
        self.simulator = simulator

    def reset(self, state: Any):
        if hasattr(self.simulator["env"].env, self.attribute):
            setattr(self.simulator["env"].env, self.attribute, state)
        else:
            self.backend.logwarn_once(f"{self.attribute} is not an attribute of the environment.")



import eagerx
from eagerx import register
from eagerx.engines.openai_gym.engine import GymEngine
from eagerx_tutorials.pendulum.objects import Pendulum


# This decorator registers the engine-specific implementation for the entity_id="Pendulum".
@register.engine(GymEngine, entity=Pendulum)
def gym_engine(spec: eagerx.specs.ObjectSpec, graph: eagerx.EngineGraph):
    """Engine-specific implementation (GymEngine) of the Pendulum object."""
    # Import the openai engine-specific nodes (ObservationSensor, ActionActuator, GymImage)
    from eagerx.engines.openai_gym.enginenodes import ObservationSensor, ActionActuator, GymImage

    # Import the tutorial engine-specific nodes (FloatOutput)
    from eagerx_tutorials.pendulum.engine_nodes import FloatOutput

    # Set engine-specific parameters
    spec.engine.env_id = "Pendulum-v1"

    # Create engine states that implement the registered states
    # Note: The GymEngine implementation unfortunately does not support setting the OpenAI environment state,
    #       nor does it support changing the dynamic parameters.
    #       However, you could create an Engine specifically for the Pendulum-v1 environment.
    from eagerx_tutorials.pendulum.engine_states import DummyState, SetGymAttribute

    spec.engine.states.model_state = DummyState.make()  # Use dummy state, so it can still be selected.
    spec.engine.states.model_parameters = DummyState.make()  # Use dummy state (same reason as above).
    spec.engine.states.mass = SetGymAttribute.make(attribute="m")
    spec.engine.states.length = SetGymAttribute.make(attribute="l")
    spec.engine.states.max_speed = SetGymAttribute.make(attribute="max_speed")

    # Create sensor engine nodes.
    image = GymImage.make("image", rate=spec.sensors.image.rate, shape=spec.config.render_shape, process=2)
    theta = FloatOutput.make("theta", rate=spec.sensors.theta.rate, idx=0)
    # Create engine node that implements the theta_dot observation
    # START EXERCISE 1.1.a
    theta_dot = FloatOutput.make("theta_dot", rate=spec.sensors.theta_dot.rate, idx=1)
    # END EXERCISE 1.1.a

    # Create actuator engine node
    action = ActionActuator.make("action", rate=spec.actuators.u.rate, process=2, zero_action=[0])

    # Use the observations produced by the "Pendulum-v1" to obtain theta and theta_dot.
    # Because this observation is [sin(theta), cos(theta), theta_dot], so we first convert it to [theta, theta_dot]
    from eagerx_tutorials.pendulum.processor import ObsWithDecomposedAngle

    x = ObservationSensor.make("x", rate=spec.sensors.theta.rate, process=2)
    x.outputs.observation.processor = ObsWithDecomposedAngle.make(convert_to="theta_theta_dot")

    # Add all engine nodes to the engine-specific graph
    graph.add([x, theta, image, action])
    # START EXERCISE 1.1.b
    graph.add(theta_dot)
    # END EXERCISE 1.1.b

    # theta
    graph.connect(source=x.outputs.observation, target=theta.inputs.observation_array)
    graph.connect(source=theta.outputs.observation, sensor="theta")

    # theta_dot
    # START EXERCISE 1.1.c
    graph.connect(source=x.outputs.observation, target=theta_dot.inputs.observation_array)
    graph.connect(source=theta_dot.outputs.observation, sensor="theta_dot")
    # END EXERCISE 1.1.c

    # image
    graph.connect(source=image.outputs.image, sensor="image")

    # u
    # Note: not to be confused with sensor "u", for which we do not provide an implementation here.
    # Note: We add a processor that negates the action, as the torque in OpenAI gym is defined counter-clockwise.
    from eagerx_tutorials.pendulum.processor import VoltageToMotorTorque

    action.inputs.action.processor = VoltageToMotorTorque.make(K=0.03333, R=7.731)
    graph.connect(actuator="u", target=action.inputs.action)



"""
CPG in polar coordinates based on:
Pattern generators with sensory feedback for the control of quadruped
authors: L. Righetti, A. Ijspeert
https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=4543306
Original author: Guillaume Bellegarda
"""

from typing import Optional
import numpy as np
import eagerx
from eagerx import Node, Space
from eagerx import register
from eagerx.utils.utils import Msg

from eagerx_tutorials.quadruped.hopf_network import HopfNetwork
import eagerx_tutorials.quadruped.go1.configs_go1 as go1_config


class CpgGait(Node):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        gait: str,
        omega_swing: float,
        omega_stance: float,
        ground_clearance: float = 0.04,
        ground_penetration: float = 0.02,
        mu: int = 2,
        couple: bool = True,
        coupling_strength: float = 1.0,
        robot_height: float = 0.25,
        des_step_len: float = 0.04,
        process: Optional[int] = eagerx.ENVIRONMENT,
    ) -> eagerx.specs.NodeSpec:
        """Make a spec to create a CpgGait node that produces a quadruped gait.

        It uses a CPG network based on hopf polar equations mapped to foot positions in Cartesian space.

        :param name: User specified node name.
        :param rate: Rate (Hz) at which the callback is called.
        :param gait: Change depending on desired gait.
        :param omega_swing: todo: MUST EDIT
        :param omega_stance:  todo: MUST EDIT
        :param ground_clearance: Foot swing height.
        :param ground_penetration: Foot stance penetration into ground.
        :param mu: todo: 1**2, converge to sqrt(mu)
        :param couple: Should couple.
        :param coupling_strength: Coefficient to multiply coupling matrix.
        :param robot_height: In nominal case (standing).
        :param des_step_len: Desired step length.
        :param process: Process in which this node is launched. See :class:`~eagerx.core.constants.process` for all options.
        :return: NodeSpec
        """
        spec = cls.get_specification()

        # Modify default params
        spec.config.update(name=name, rate=rate, process=process, inputs=["offset"], outputs=["cartesian_pos", "xs_zs"])

        # Modify params (args to .initialize())
        spec.config.update(mu=mu, gait=gait, omega_swing=omega_swing, omega_stance=omega_stance)
        spec.config.update(ground_clearance=ground_clearance, ground_penetration=ground_penetration)
        spec.config.update(couple=couple, coupling_strength=coupling_strength)
        spec.config.update(robot_height=robot_height, des_step_len=des_step_len)
        return spec

    def initialize(self, spec):
        assert spec.config.gait == "TROT", "xs_zs is only correct for TROT gait."
        self.n_legs = 4
        self.side_sign = np.array([-1, 1, -1, 1])  # get correct hip sign (body right is negative)
        self.foot_y = 0.0838  # this is the hip length
        self.cpg = HopfNetwork(
            mu=spec.config.mu,
            gait=spec.config.gait,
            omega_swing=spec.config.omega_swing,
            omega_stance=spec.config.omega_stance,
            time_step=0.005,  # Always update cpg with 200 Hz.
            ground_clearance=spec.config.ground_clearance,  # foot swing height
            ground_penetration=spec.config.ground_penetration,  # foot stance penetration into ground
            robot_height=spec.config.robot_height,  # in nominal case (standing)
            des_step_len=spec.config.des_step_len,  # 0 for jumping
        )

    @register.states()
    def reset(self):
        self.cpg.reset()

    @register.inputs(offset=Space(low=[-0.01] * 4, high=[0.01] * 4))
    @register.outputs(
        cartesian_pos=Space(shape=(len(go1_config.NOMINAL_FOOT_POS_LEG_FRAME),), dtype="float32"),  # TODO: Set correct bounds
        xs_zs=Space(
            low=[-0.05656145, -0.26999995, -0.05656852, -0.2699973], high=[0.05636625, -0.21000053, 0.05642071, -0.21001561]
        ),
    )
    def callback(self, t_n: float, offset: Msg):
        # update CPG
        while self.cpg.t <= t_n:
            self.cpg.update()

        # get desired foot positions from CPG
        xs, zs = self.cpg.get_xs_zs()

        # get unique xs & zs positions (BASED ON TROT)
        unique_xs_zs = np.array([xs[0], zs[0], xs[1], zs[1]], dtype="float32")

        action = np.zeros((12,), dtype="float32")
        offset = offset.msgs[-1].data
        for i in range(self.n_legs):
            xyz_desired = np.array([xs[i], self.side_sign[i] * self.foot_y + offset[i], zs[i]])
            action[3 * i : 3 * i + 3] = xyz_desired
        return dict(cartesian_pos=action, xs_zs=unique_xs_zs)



"""
CPG in polar coordinates based on:
Pattern generators with sensory feedback for the control of quadruped
authors: L. Righetti, A. Ijspeert
https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=4543306
Original author: Guillaume Bellegarda
"""

from functools import partial
import numpy as np


# Runge Kutta taking the EOM as an argument
def runge_kutta4(x, ode, dt):
    k1 = ode(x)
    k2 = ode(x + 0.5 * dt * k1)
    k3 = ode(x + 0.5 * dt * k2)
    k4 = ode(x + dt * k3)
    return x + (dt / 6) * (k1 + 2 * k2 + 2 * k3 + k4)


# @jit(nopython=True)
def ode(omega_swing, omega_stance, mu, alpha, couple, coupling_strength, PHI, x):
    r = x[:4]
    theta = x[4:]

    # Calculate r_dot
    dr = alpha * r * (mu - r**2)

    # Calculate theta dot
    theta_dot = np.array([0, 0, 0, 0], dtype="float32")
    mask = np.sin(theta) > 0
    theta_dot[mask] = omega_swing
    theta_dot[~mask] = omega_stance

    # Calculate coupling
    if couple:
        for i in range(4):
            for j in range(4):
                if j != i:
                    theta_dot[i] += r[j] * coupling_strength * np.sin(theta[j] - theta[i] - PHI[i, j])
    return np.array([dr[0], dr[1], dr[2], dr[3], theta_dot[0], theta_dot[1], theta_dot[2], theta_dot[3]], dtype="float32")


class HopfNetwork:
    """CPG network based on hopf polar equations mapped to foot positions in Cartesian space.

    Foot Order is FR, FL, RR, RL
    (Front Right, Front Left, Rear Right, Rear Left)
    """

    def __init__(
        self,
        mu=1,  # 1**2,            # converge to sqrt(mu)
        omega_swing=1 * 2 * np.pi,  # MUST EDIT
        omega_stance=1 * 2 * np.pi,  # MUST EDIT
        gait="TROT",  # change depending on desired gait
        coupling_strength=1,  # coefficient to multiply coupling matrix
        couple=True,  # should couple
        time_step=0.001,  # time step
        ground_clearance=0.05,  # foot swing height
        ground_penetration=0.01,  # foot stance penetration into ground
        robot_height=0.25,  # in nominal case (standing)
        des_step_len=0.04,  # desired step length
    ):

        ###############
        # initialize CPG data structures: amplitude is row 0, and phase is row 1
        self.X = np.zeros((2, 4), dtype="float32")

        # save parameters
        self._mu = mu
        self._omega_swing = omega_swing
        self._omega_stance = omega_stance
        self._couple = couple
        self._coupling_strength = coupling_strength
        self._dt = time_step
        self.t = None
        self._set_gait(gait)

        # save body and foot shaping
        self._ground_clearance = ground_clearance
        self._ground_penetration = ground_penetration
        self._robot_height = robot_height
        self._des_step_len = des_step_len

        self.ode = partial(
            ode, self._omega_swing, self._omega_stance, self._mu, 50, self._couple, self._coupling_strength, self.PHI
        )

    def _set_gait(self, gait):
        """For coupling oscillators in phase space."""
        self.PHI_trot = np.pi * np.array(
            [
                [0, -1, -1, 0],
                [1, 0, 0, 1],
                [1, 0, 0, 1],
                [0, -1, -1, 0],
            ]
        )

        self.PHI_jump = np.pi * np.array(
            [
                [1, 1, 1, 1],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
            ]
        )

        self.PHI_walk = np.pi * np.array(
            [
                [0, -1, -1 / 2, 1 / 2],
                [1, 0, 1 / 2, 3 / 2],
                [1 / 2, -1 / 2, 0, 1],
                [-1 / 2, -3 / 2, -1, 0],
            ]
        )

        self.PHI_bound = np.pi * np.array(
            [
                [0, 0, -1, -1],
                [0, 0, -1, -1],
                [1, 1, 0, 0],
                [1, 1, 0, 0],
            ]
        )

        self.PHI_pace = np.pi * np.array(
            [
                [0, -1, 0, -1],
                [1, 0, 1, 0],
                [0, -1, 0, -1],
                [1, 0, 1, 0],
            ]
        )

        self.PHI = {
            "TROT": self.PHI_trot,
            "PACE": self.PHI_pace,
            "BOUND": self.PHI_bound,
            "WALK": self.PHI_walk,
            "JUMP": self.PHI_jump,
        }[gait]

        # print(gait)

    def reset(self):
        # reset current time
        self.t = 0
        # set oscillator initial conditions
        self.X[0, :] = 0.1
        # initialize so we are in stance phase for the trot gait
        self.X[1, :] = (self.PHI[0, :] + 4 / 3 * np.pi) % (2 * np.pi)

    def get_xs_zs(self):
        # map CPG variables to Cartesian foot xz positions (Equations 8, 9)
        r = self.X[0, :]
        theta = self.X[1, :]
        x = -self._des_step_len * r * np.cos(theta)

        ground_clearance = self._ground_clearance * np.sin(theta)
        ground_penetration = self._ground_penetration * np.sin(theta)
        above_ground = np.sin(theta) > 0
        ground_offset = above_ground * ground_clearance + (1.0 - above_ground) * ground_penetration
        z = -self._robot_height + ground_offset
        return x, z

    def update(self):
        """Update oscillator states."""
        # update parameters, integrate
        # self._integrate_hopf_equations()

        # Predict with rk4
        x = self.X.reshape((-1,)).copy()
        x = runge_kutta4(x, self.ode, self._dt)
        X = x.reshape(2, 4)
        X[1, :] = X[1, :] % (2 * np.pi)
        self.X = X

        # move forward time
        self.t += self._dt

    def _integrate_hopf_equations(self):
        """Hopf polar equations and integration. Use equations 6 and 7."""
        # bookkeeping - save copies of current CPG states
        X = self.X.copy()
        X_dot = np.zeros((2, 4))
        alpha = 50

        # loop through each leg's oscillator
        for i in range(4):
            # get r_i, theta_i from X
            r = X[0, i]
            theta = X[1, i]
            # compute r_dot (Equation 6)
            r_dot = alpha * (self._mu - r**2) * r
            # determine whether oscillator i is in swing or stance phase to set natural frequency omega_swing or omega_stance (see Section 3)
            theta_dot = 0.0
            if np.sin(theta) > 0:
                theta_dot = self._omega_swing
            else:
                theta_dot = self._omega_stance

            # loop through other oscillators to add coupling (Equation 7)
            if self._couple:
                for j in range(4):
                    if j != i:
                        theta_dot += X[0, j] * self._coupling_strength * np.sin(X[1, j] - theta - self.PHI[i, j])

            # set X_dot[:,i]
            X_dot[:, i] = [r_dot, theta_dot]

        # integrate
        self.X += self._dt * X_dot
        # mod phase variables to keep between 0 and 2pi
        self.X[1, :] = self.X[1, :] % (2 * np.pi)

        # self.X_list.append(self.X.copy())
        # self.dX_list.append(X_dot)



from typing import List

# EAGERx IMPORTS
from eagerx_ode.engine import OdeEngine
from eagerx import Object, Space
from eagerx.core.specs import ObjectSpec
from eagerx.core.graph_engine import EngineGraph
import eagerx.core.register as register


class Pendulum(Object):
    @classmethod
    @register.sensors(
        theta=Space(low=-999.0, high=999.0, shape=(), dtype="float32"),
        theta_dot=Space(low=-999.0, high=999.0, shape=(), dtype="float32"),
        image=Space(dtype="uint8"),  # shape, low & high determined at run-time
        u_applied=Space(low=[-2.0], high=[2.0], dtype="float32"),
    )
    @register.actuators(u=Space(low=[-2], high=[2], dtype="float32"))
    @register.engine_states(
        model_state=Space(low=[-3.14, -9], high=[3.14, 9], dtype="float32"),
        model_parameters=Space(dtype="float32"),  # shape, low & high determined at run-time
        mass=Space(low=0.01, high=0.07, shape=(), dtype="float32"),
        length=Space(low=0.04, high=0.2, shape=(), dtype="float32"),
        max_speed=Space(low=22, high=22, shape=(), dtype="float32"),
    )
    def make(
        cls,
        name: str,
        actuators: List[str] = None,
        sensors: List[str] = None,
        states: List[str] = None,
        rate: float = 30.0,
        render_shape: List[int] = None,
        render_fn: str = None,
    ):
        """Make the specification of the Pendulum.

        Sensors
        theta: angle of the pendulum wrt upward position
        theta_dot: angular velocity of the pendulum
        image: render of pendulum system
        u_applied: Applied DC motor voltage

        Actuators
        u: DC motor voltage

        States
        model_state: allows resetting the angle and angular velocity
        model_parameters: allows resetting all ODE parameters [J, m, l, b, K, R, c, d].
        mass: allows resetting the mass of the Gym pendulum m
        length: allows resetting the length of the Gym pendulum l
        max_speed: allows resetting the max speed of the Gym pendulum

        Config
        render_shape: shape of render window [height, width]
        """
        spec = cls.get_specification()

        # Modify default agnostic params
        # Only allow changes to the agnostic params (rates, windows, (space)converters, etc...
        spec.config.name = name
        spec.config.sensors = ["theta", "theta_dot"] if sensors is None else sensors
        spec.config.actuators = ["u"] if actuators is None else actuators
        spec.config.states = ["model_state"] if states is None else states

        # Add custom agnostic params
        spec.config.render_shape = render_shape if render_shape else [480, 480]
        spec.config.render_fn = render_fn if render_fn else "pendulum_render_fn"

        # Set observation properties: (space_converters, rate, etc...)
        spec.sensors.theta.rate = rate
        spec.sensors.theta_dot.rate = rate
        spec.sensors.image.rate = rate / 2
        spec.sensors.u_applied.rate = rate
        spec.actuators.u.rate = rate

        # Set image space
        shape = (spec.config.render_shape[0], spec.config.render_shape[1], 3)
        spec.sensors.image.space = Space(low=0, high=255, shape=shape, dtype="uint8")

        # Set model_parameters properties: (space_converters)
        # Set default params of pendulum ode [J, m, l, b, K, R, c, d].
        mean = [
            0.000159931461600856,
            0.0508581731919534,
            0.0415233722862552,
            1.43298488358436e-05,
            0.0333391179016334,
            7.73125142447252,
            0.000975041213361349,
            165.417960777425,
        ]
        diff = [0, 0, 0, 0, 0, 0, 0, 0]  # Percentual delta with respect to fixed value
        low = [val - diff * val for val, diff in zip(mean, diff)]
        high = [val + diff * val for val, diff in zip(mean, diff)]
        spec.states.model_parameters.space = Space(low=low, high=high, dtype="float32")

        return spec

    @staticmethod
    @register.engine(OdeEngine)  # This decorator pre-initializes engine implementation with default object_params
    def ode_engine(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation (OdeEngine) of the object."""
        # Set object arguments
        spec.engine.ode = "eagerx_tutorials.pendulum.pendulum_ode/pendulum_ode"
        spec.engine.Dfun = "eagerx_tutorials.pendulum.pendulum_ode/pendulum_dfun"

        # Set default params of pendulum ode [J, m, l, b, K, R, c, d].
        spec.engine.ode_params = [
            0.000159931461600856,
            0.0508581731919534,
            0.0415233722862552,
            1.43298488358436e-05,
            0.0333391179016334,
            7.73125142447252,
            0.000975041213361349,
            165.417960777425,
        ]

        # Create engine_states (no agnostic states defined in this case)
        from eagerx_ode.engine_states import OdeEngineState, OdeParameters
        from eagerx_tutorials.pendulum.engine_states import DummyState
        from eagerx_tutorials.pendulum.engine_nodes import FloatOutput
        from eagerx_ode.engine_nodes import OdeOutput, OdeInput, OdeRender, ActionApplied

        spec.engine.states.model_state = OdeEngineState.make()
        spec.engine.states.model_parameters = OdeParameters.make(list(range(8)))
        spec.engine.states.mass = DummyState.make()
        spec.engine.states.length = DummyState.make()
        spec.engine.states.max_speed = DummyState.make()

        # Create sensor engine nodes
        x = OdeOutput.make("x", rate=spec.sensors.theta.rate, process=2)

        # For didactic purposes, we create two sensors, i.e. one with angle and one with angular velocity.
        # We could also have created a sensor that contains both, but in this way it is more clear which sensor
        # contains what information.
        theta = FloatOutput.make("theta", rate=spec.sensors.theta.rate, idx=0)
        theta_dot = FloatOutput.make("theta_dot", rate=spec.sensors.theta_dot.rate, idx=1)

        u_applied = ActionApplied.make("u_applied", rate=spec.sensors.u_applied.rate, process=2)

        render_fn = f"eagerx_tutorials.pendulum.pendulum_render/{spec.config.render_fn}"
        shape = spec.sensors.image.space.shape[:2]
        image = OdeRender.make("image", render_fn=render_fn, rate=spec.sensors.image.rate, process=2, shape=shape)

        # Create actuator engine nodes
        u = OdeInput.make("u", rate=spec.actuators.u.rate, process=2, default_action=[0])

        # Connect all engine nodes
        graph.add([x, theta, theta_dot, image, u, u_applied])

        # theta
        graph.connect(source=x.outputs.observation, target=theta.inputs.observation_array)
        graph.connect(source=theta.outputs.observation, sensor="theta")

        # theta_dot
        graph.connect(source=x.outputs.observation, target=theta_dot.inputs.observation_array)
        graph.connect(source=theta_dot.outputs.observation, sensor="theta_dot")

        # image
        graph.connect(source=x.outputs.observation, target=image.inputs.observation)
        graph.connect(source=image.outputs.image, sensor="image")

        # u
        graph.connect(actuator="u", target=u.inputs.action)

        graph.connect(source=u.outputs.action_applied, target=image.inputs.action_applied, skip=True)
        graph.connect(source=u.outputs.action_applied, target=u_applied.inputs.action_applied, skip=True)
        graph.connect(source=u_applied.outputs.action_applied, sensor="u_applied")

        # Check graph validity (commented out)
        # graph.is_valid(plot=True)



from collections import deque


class PID:
    def __init__(self, u0: float, kp: float, kd: float, ki: float, dt: float):
        self.u0 = u0
        self.kp = kp
        self.kd = kd
        self.ki = ki
        self.dt = dt

        self.F = [kp + ki * dt + kd / dt, -kp - 2 * kd / dt, kd / dt]

        self.window = None
        self.u = None

    def reset(self):
        self.u = self.u0
        self.window = deque(maxlen=3)

    def next_action(self, y: float, ref: float = 0.0):
        # Add error
        self.window.appendleft(ref - y)

        # Calculate action
        for idx, e in enumerate(self.window):
            self.u += self.F[idx] * e

        return self.u



import eagerx
from eagerx import register, Space
from eagerx.utils.utils import Msg
import cv2
import numpy as np


class Overlay(eagerx.Node):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: int = eagerx.process.ENVIRONMENT,
        color: str = "cyan",
    ):
        """Overlay spec"""
        spec = cls.get_specification()

        # Adjust default params
        spec.config.update(
            name=name, rate=rate, process=process, color=color, inputs=["base_image", "u", "theta"], outputs=["image"]
        )
        return spec

    def initialize(self, spec):
        pass

    @register.states()
    def reset(self):
        pass

    @register.inputs(
        base_image=Space(dtype="uint8"),
        u=Space(low=[-3], high=[3]),
        theta=Space(shape=(), dtype="float32"),
    )
    @register.outputs(image=Space(dtype="uint8"))
    def callback(self, t_n: float, base_image: Msg, u: Msg, theta: Msg):
        if len(base_image.msgs[-1].data) > 0:
            u = u.msgs[-1].data[0] if u else 0
            theta = theta.msgs[-1]

            # Set background image from base_image
            img = base_image.msgs[-1]
            height, width, _ = img.shape
            side_length = min(width, height)

            # Put text
            font = cv2.FONT_HERSHEY_SIMPLEX
            text = "Applied Voltage"
            text_size = cv2.getTextSize(text, font, 0.5, 2)[0]
            text_x = int((width - text_size[0]) / 2)
            text_y = int(text_size[1])
            img = cv2.putText(img, text, (text_x, text_y), font, 0.5, (0, 0, 0))

            # Draw grey bar
            img = cv2.rectangle(
                img,
                (width // 2 - side_length * 4 // 10, height // 2 - side_length * 9 // 20),
                (width // 2 + 4 * side_length // 10, height // 2 - 4 * side_length // 10),
                (125, 125, 125),
                -1,
            )

            # Fill bar proportional to the action that is applied
            p1 = (width // 2, height // 2 - side_length * 9 // 20)
            p2 = (width // 2 + int(side_length * u * 2 / 15), height // 2 - 4 * side_length // 10)
            img = cv2.rectangle(img, p1, p2, (0, 0, 0), -1)

            # START EXERCISE 1.3
            img = cv2.putText(img, f"t ={t_n: .2f} s", (text_x, height - text_y), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0))
            # START EXERCISE 1.3

            # Add theta info
            img = cv2.putText(
                img, f"theta ={theta: .2f} rad", (text_x, height - int(2.2 * text_y)), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0)
            )

            return dict(image=img)
        else:
            return dict(image=np.zeros((0, 0, 3), dtype="uint8"))



import eagerx
from eagerx import register, Space
from eagerx.utils.utils import Msg
from typing import List
import cv2
from matplotlib.cm import get_cmap
import numpy as np
from collections import deque
from functools import partial


class XyPlane(eagerx.Node):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: int = eagerx.ENVIRONMENT,
        color: str = "cyan",
        px_pm: int = 45,
        top_left: List[int] = None,
        lower_right: List[int] = None,
        num_eps: int = 20,
        colormap: str = "plasma",
    ) -> eagerx.specs.NodeSpec:
        """XyPlane spec"""
        spec = cls.get_specification()

        # Adjust default params
        spec.config.update(name=name, rate=rate, process=process, color=color, inputs=["position"], outputs=["image"])
        spec.config.px_pm = px_pm
        spec.config.colormap = colormap
        spec.config.top_left = top_left if isinstance(top_left, list) else [-3, -3]  # x,y in [m]
        spec.config.lower_right = lower_right if isinstance(lower_right, list) else [6, 8]  # x,y in [m]
        spec.config.num_eps = num_eps
        spec.inputs.position.window = 0  # Receive all new position messages since last callback

        # Calculate shape
        tl_x, tl_y = spec.config.top_left
        lr_x, lr_y = spec.config.lower_right
        width = int((lr_y - tl_y) * spec.config.px_pm)  # px
        height = int((lr_x - tl_x) * spec.config.px_pm)
        spec.outputs.image.space = Space(low=0, high=255, shape=(height, width, 3), dtype="uint8")
        return spec

    def initialize(self, spec):
        # Calculate width, height, and shape
        self.px_pm = spec.config.px_pm
        self.thickness = max(int(spec.config.px_pm * 0.10), 1)
        self.top_left = spec.config.top_left
        self.lower_right = spec.config.lower_right
        tl_x, tl_y = spec.config.top_left
        lr_x, lr_y = spec.config.lower_right
        self.width = int((lr_y - tl_y) * spec.config.px_pm)  # px
        self.height = int((lr_x - tl_x) * spec.config.px_pm)
        self.shape = (self.height, self.width)

        # Prepare history
        self.num_eps = spec.config.num_eps
        self.last_xy = deque(maxlen=spec.config.num_eps)
        self.colors = deque(maxlen=spec.config.num_eps)
        self.xy = None  # [m]

        # Prepare colormap
        self.cmap = get_cmap(spec.config.colormap)
        interp = np.linspace(0, 1, num=spec.config.num_eps + 1, dtype="float32")
        self.colormap = [self.cmap(i, bytes=True) for i in reversed(interp)]
        self.colormap = [(int(i[2]), int(i[1]), int(i[0])) for i in self.colormap]

        # Plot points
        self.plot_px = partial(
            self._plot_px,
            thickness=self.thickness,
            height=self.height,
            width=self.width,
            px_pm=spec.config.px_pm,
            tl_x=tl_x,
            tl_y=tl_y,
        )

    @register.states()
    def reset(self):
        # Add xy-coordinates from previous episode to buffer
        if self.xy is None:  # First reset
            self.xy = []
        if len(self.xy) > 1:
            if len(self.last_xy) == self.num_eps:
                self.colors = deque([i - 1 for i in self.colors], maxlen=self.num_eps)
                self.colors.append(len(self.last_xy) - 1)
            else:
                self.colors.append(len(self.last_xy))
            self.last_xy.append(self.xy)

        # Empty xy-coordinates
        self.xy = []  # [m]

        # Construct base image with xy-coordinates from previous episodes
        self.base_img = 255 * np.ones((self.height, self.width, 3), dtype="uint8")
        tl_x, tl_y = self.top_left
        lr_x, lr_y = self.lower_right
        self._plot_overlay(self.base_img, (0, 0, 0), 3, self.width, self.px_pm, tl_x, tl_y, lr_x, lr_y)
        for idx, xy_list in enumerate(self.last_xy):
            color = self.colormap[self.colors[idx]]
            # p = (self.num_eps - len(self.last_xy) + idx) / self.num_eps
            # opaque = tuple([int(255 - p * (255 - x)) for x in color])
            for xy in xy_list:
                self.plot_px(self.base_img, xy, color)

    @staticmethod
    def _plot_overlay(img, color, thickness, width, px_pm, tl_x, tl_y, lr_x, lr_y):
        # Put text
        font = cv2.FONT_HERSHEY_SIMPLEX
        text = "Top-view (i.e. xy-plane)"
        text_size = cv2.getTextSize(text, font, 0.75, 2)[0]
        text_x = int((width - text_size[0]) / 2)
        text_y = int(text_size[1])
        img = cv2.putText(img, text, (text_x, text_y), font, 0.75, thickness=2, color=(0, 0, 0))

        center = (int(-tl_y * px_pm), int(-tl_x * px_pm))

        # x-axis
        x = center[0] + int(lr_y * px_pm * 0.75)
        cv2.arrowedLine(img, center, (x, center[1]), color, thickness, tipLength=0.1)
        text_x = (x + text_size[1], center[1] + text_size[1] // 2)
        cv2.putText(img, "x", text_x, font, 0.75, thickness=2, color=(0, 0, 0))
        # y-axis
        y = center[1] + int(lr_x * px_pm * 0.75)
        cv2.arrowedLine(img, center, (center[0], y), color, thickness, tipLength=0.1)
        text_y = (center[0] - text_size[1] // 2, y + text_size[1])
        cv2.putText(img, "y", text_y, font, 0.75, thickness=2, color=(0, 0, 0))

    @staticmethod
    def _plot_px(img, xy, color, thickness, height, width, px_pm, tl_x, tl_y):
        x, y = xy
        x -= tl_x
        y -= tl_y
        px_x = int(x * px_pm)
        px_y = int(y * px_pm)
        if px_y > height or px_x > width:
            return
        # Plot pixel
        cv2.circle(img, (px_y, px_x), thickness, color, -1)

    @register.inputs(position=Space(low=[-10.0, -10.0, 0.0], high=[10.0, 10.0, 2.0]))
    @register.outputs(image=Space(dtype="uint8"))
    def callback(self, t_n: float, position: Msg):
        # Select newest color
        idx = len(self.last_xy)
        color = self.colormap[idx]

        # plot points
        for xyz in position.msgs:
            xy = [xyz.data[0], xyz.data[1]]
            self.plot_px(self.base_img, xy, color)
            self.xy.append(xy)

        # Prepare image for transmission.
        return dict(image=self.base_img)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eagerx_tutorials',
 'eagerx_tutorials.pendulum',
 'eagerx_tutorials.quadruped',
 'eagerx_tutorials.quadruped.go1']

package_data = \
{'': ['*'],
 'eagerx_tutorials.quadruped.go1': ['go1_description/meshes/*',
                                    'go1_description/urdf/*']}

install_requires = \
['PyVirtualDisplay>=3.0,<4.0',
 'eagerx-ode>=0.1.26,<0.2.0',
 'eagerx-pybullet>=0.1.10,<0.2.0',
 'eagerx>=0.1.33,<0.2.0',
 'ipywidgets>=7.7.0,<8.0.0',
 'jupyterlab>=3.3.4,<4.0.0',
 'nbconvert>=6.5.0,<7.0.0',
 'sb3-contrib>=1.5.0,<2.0.0',
 'stable-baselines3>=1.2.0,<2.0.0',
 'tqdm>=4.64.0,<5.0.0']

setup_kwargs = {
    'name': 'eagerx-tutorials',
    'version': '0.1.24',
    'description': 'Tutorials on how to use EAGERx.',
    'long_description': 'None',
    'author': 'Jelle Luijkx',
    'author_email': 'j.d.luijkx@tudelft.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/eager-dev/eagerx_tutorials',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
